<?php

namespace Drupal\inspector_monitoring\Subscribers;

use Drupal\Core\Path\PathMatcherInterface;
use Drupal\inspector_monitoring\Service\Inspector;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Event\ControllerArgumentsEvent;
use Symfony\Component\HttpKernel\Event\ControllerEvent;
use Symfony\Component\HttpKernel\Event\ExceptionEvent;
use Symfony\Component\HttpKernel\Event\FinishRequestEvent;
use Symfony\Component\HttpKernel\Event\KernelEvent;
use Symfony\Component\HttpKernel\Event\RequestEvent;
use Symfony\Component\HttpKernel\Event\ResponseEvent;
use Symfony\Component\HttpKernel\Event\TerminateEvent;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\Routing\Route;

/**
 * Kernel events subscriber.
 */
class KernelEventsSubscriber implements EventSubscriberInterface {
  use InspectorAwareTrait;

  /**
   * The segment type for controllers.
   */
  protected const SEGMENT_TYPE_CONTROLLER = 'controller';

  /**
   * The segment type for processes.
   */
  protected const SEGMENT_TYPE_PROCESS = 'process';

  /**
   * The segment type for templates.
   */
  protected const SEGMENT_TYPE_TEMPLATE = 'template';

  /**
   * The route path.
   */
  protected string $routePath;

  /**
   * Constructs a new KernelEventsSubscriber object.
   *
   * @param \Drupal\inspector_monitoring\Service\Inspector $inspector
   *   The inspector service.
   * @param \Drupal\Core\Path\PathMatcherInterface $pathMatcher
   *   The path matcher service.
   */
  public function __construct(
    protected Inspector $inspector,
    protected PathMatcherInterface $pathMatcher,
  ) {
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    // The higher the priority number, the earlier the method is called.
    return [
      KernelEvents::CONTROLLER => ['onKernelController', 9999],
      KernelEvents::CONTROLLER_ARGUMENTS => [
        ['onKernelPreControllerArguments', 9999],
        ['onKernelPostControllerArguments', -9999],
      ],
      KernelEvents::EXCEPTION => ['onKernelException', 9999],
      KernelEvents::FINISH_REQUEST => ['onKernelFinishRequest', 9999],
        // The priority set AFTER the authentication subscriber to ensure the
        // inspector transaction is already initialized.
      KernelEvents::REQUEST => [['onKernelRequest', 400]/*, ['setAuthUserInfo', 30]*/],
      KernelEvents::RESPONSE => ['onKernelResponse', 9999],
      KernelEvents::VIEW => ['onKernelView', 9999],
      KernelEvents::TERMINATE => ['onKernelTerminate', -9999],

    ];
  }

  /**
   * Intercept a controller event.
   *
   * @param \Symfony\Component\HttpKernel\Event\ControllerEvent $event
   *   The controller event.
   */
  public function onKernelController(ControllerEvent $event): void {
    if (!$this->inspector->canAddSegments()) {
      return;
    }

    $this->endSegment(KernelEvents::REQUEST);

    $this->startSegment(self::SEGMENT_TYPE_CONTROLLER, KernelEvents::CONTROLLER)
      ->addContext("Controller", ['class' => $event->getController()]);
  }

  /**
   * Intercept a controller arguments event.
   *
   * @param \Symfony\Component\HttpKernel\Event\ControllerArgumentsEvent $event
   *   The controller arguments event.
   */
  public function onKernelPreControllerArguments(ControllerArgumentsEvent $event): void {
    if (!$this->inspector->canAddSegments()) {
      return;
    }

    $this->endSegment(KernelEvents::CONTROLLER);

    $this->startSegment(self::SEGMENT_TYPE_CONTROLLER, KernelEvents::CONTROLLER_ARGUMENTS)
      ->addContext("Controller", ['class' => $event->getController(), 'arguments' => $event->getArguments()]);
  }

  /**
   * Intercept a controller arguments event.
   *
   * @param \Symfony\Component\HttpKernel\Event\ControllerArgumentsEvent $event
   *   The controller arguments event.
   */
  public function onKernelPostControllerArguments(ControllerArgumentsEvent $event): void {
    if (!$this->inspector->canAddSegments()) {
      return;
    }

    $this->endSegment(KernelEvents::CONTROLLER_ARGUMENTS);

    $controllerLabel = $this->controllerLabel($event);

    $arguments = [];
    foreach ($event->getArguments() as $argument) {
      if (\is_object($argument)) {
        $args = ['class' => \get_class($argument)];
        if (\method_exists($argument, 'getId')) {
          $args['id'] = $argument->getId();
        }
        $arguments[] = $args;
      }
      else {
        $arguments[] = $argument;
      }
    }

    $this->startSegment(self::SEGMENT_TYPE_CONTROLLER, $controllerLabel)
      ->addContext('Data', ['arguments' => $arguments]);
  }

  /**
   * Intercept an HTTP request.
   *
   * @param \Symfony\Component\HttpKernel\Event\RequestEvent $event
   *   The request event.
   *
   * @throws \Exception
   */
  public function onKernelRequest(RequestEvent $event): void {
    if (!$this->inspector->isRecording() || !$this->isMasterMainRequest($event)) {
      return;
    }

    $request = $event->getRequest();

    // Retrieve the route pattern.
    try {

      /** @var \Symfony\Component\Routing\Route $route */
      $route = $request->attributes->get('_route_object');

      // If the route is not in the route collection, $route could be null.
      if ($route instanceof Route) {
        $this->routePath = $route->getPath();
      }
      else {
        $this->routePath = $request->getPathInfo();
      }

    }
    catch (\Throwable $exception) {
      $this->routePath = $request->getPathInfo();
    }

    if (!$this->isRequestEligibleForInspection($this->routePath)) {
      return;
    }

    $this->startTransaction(
      $event->getRequest()->getMethod() . ' ' . \trim($this->routePath, '/'))
      ->markAsRequest();

    $this->startSegment(self::SEGMENT_TYPE_PROCESS, KernelEvents::REQUEST);
  }

  /**
   * Ending transaction.
   *
   * @param \Symfony\Component\HttpKernel\Event\ResponseEvent $event
   *   The response event.
   */
  public function onKernelResponse(ResponseEvent $event): void {
    if (!$this->inspector->canAddSegments() || !$this->isMasterMainRequest($event)) {
      return;
    }

    // End segment based on a kernel event type.
    $controllerLabel = $this->controllerLabel($event);
    if ($controllerLabel) {
      $this->endSegment($controllerLabel);
    }
    $this->endSegment(KernelEvents::REQUEST);
    $this->endSegment(KernelEvents::VIEW);

    $response = $event->getResponse();

    $segment = $this->startSegment(self::SEGMENT_TYPE_PROCESS, KernelEvents::RESPONSE);
    $segment->addContext('Response', [
      'headers' => $response->headers->all(),
      'protocolVersion' => $response->getProtocolVersion(),
      'statusCode' => $response->getStatusCode(),
      'charset' => $response->getCharset(),
    ]);
  }

  /**
   * Intercept a finish request event.
   *
   * @param \Symfony\Component\HttpKernel\Event\FinishRequestEvent $event
   *   The finish request event.
   */
  public function onKernelFinishRequest(FinishRequestEvent $event): void {
    if ($this->inspector->canAddSegments() && $this->isMasterMainRequest($event)) {
      $this->endSegment(KernelEvents::RESPONSE);
    }
  }

  /**
   * Handle an http kernel exception.
   *
   * @param \Symfony\Component\HttpKernel\Event\ExceptionEvent $event
   *   The exception event.
   *
   * @throws \Exception
   */
  public function onKernelException($event): void {
    if (!$this->inspector->isRecording()) {
      return;
    }

    // Compatibility with Symfony < 5 and Symfony >=5
    // The additional `method_exists` check is to prevent errors in Symfony 4.3
    // where the ExceptionEvent exists and is used but doesn't implement
    // the `getThrowable` method, which was introduced in Symfony 4.4.
    if ($event instanceof ExceptionEvent && \method_exists($event, 'getThrowable')) {
      $exception = $event->getThrowable();
    }
    else {
      throw new \LogicException('Invalid exception event.');
    }

    $this->inspector->reportException($exception, FALSE);
  }

  /**
   * Intercept a terminate event.
   *
   * @param \Symfony\Component\HttpKernel\Event\TerminateEvent $event
   *   The terminate event.
   */
  public function onKernelTerminate(TerminateEvent $event): void {
    if ($this->inspector->hasTransaction()) {
      $this->inspector->transaction()->setResult($event->getResponse()->getStatusCode());
    }
  }

  /**
   * Intercept a view event.
   *
   * @param \Symfony\Component\HttpKernel\Event\ViewEvent $event
   *   The view event.
   */
  public function onKernelView(ViewEvent $event): void {
    if (!$this->inspector->canAddSegments() || $this->isMasterMainRequest($event)) {
      return;
    }

    $controllerLabel = $event->getRequest()->attributes->get('_controller');

    if ($controllerLabel) {
      $this->endSegment($controllerLabel);
    }

    $this->startSegment(self::SEGMENT_TYPE_TEMPLATE, KernelEvents::VIEW);
  }

  /**
   * Determine if the request is eligible for inspection.
   *
   * @param string $path
   *   The path of the request.
   *
   * @return bool
   *   TRUE if the request is eligible for inspection, FALSE otherwise.
   */
  protected function isRequestEligibleForInspection($path): bool {
    // @todo enable ignored_routes configuration.
    /*foreach ($this->ignoredRoutes as $pattern) {
    if (Filters::matchWithWildcard($pattern, $path)) {
    return false;
    }
    }*/

    return TRUE;
  }

  /**
   * Get the controller label.
   *
   * @param \Symfony\Component\HttpKernel\Event\KernelEvent $event
   *   The kernel event.
   *
   * @return string|null
   *   The controller label.
   */
  private function controllerLabel(KernelEvent $event): ?string {
    $controllerLabel = $event->getRequest()->attributes->get('_controller');

    if (is_null($controllerLabel)) {
      return NULL;
    }

    if (is_string($controllerLabel)) {
      return $controllerLabel;
    }

    if (is_array($controllerLabel)) {
      return implode('::', $controllerLabel);
    }

    return '_controller';
  }

  /**
   * Determine if the event is a master main request.
   *
   * @param \Symfony\Component\HttpKernel\Event\KernelEvent $event
   *   The kernel event.
   *
   * @return bool
   *   TRUE if the event is a master main request, FALSE otherwise.
   */
  private function isMasterMainRequest(KernelEvent $event): bool {
    return (method_exists($event, 'isMainRequest') && $event->isMainRequest())
      || (method_exists($event, 'isMasterRequest') && $event->isMainRequest());
  }

}
